// copyright 2022 zamfofex
// license: AGPLv3 or later

#include <mimimi/chapters.h>
#include <mimimi/senran-kagura/characters.h>
#include <mimimi/appearances.h>
#include <mimimi/allocators.h>
#include <mimimi/ground.h>
#include <mimimi/compound-behaviors.h>
#include <mimimi/geometry.h>
#include <mimimi/assets.h>
#include <mimimi/engines.h>
#include <mimimi/controls.h>
#include <mimimi/sprites.h>
#include <mimimi/cameras.h>
#include <mimimi/behaviors.h>
#include <mimimi/animations.h>
#include <mimimi/collisions.h>
#include <mimimi/poses.h>
#include <mimimi/fonts.h>
#include <mimimi/font.h>
#include <mimimi/coroutines.h>

static void mimimi_chapter_I_animation_set(struct mimimi_animation_set *set, struct mimimi_allocator *allocator)
{
	int animation_count = 8;
	int image_count = 6;
	int width = 48;
	int height = 48;
	
	int total_animation_count = 3 + animation_count;
	int total_image_count = 2 + image_count * animation_count + image_count;
	
	struct mimimi_animation *animations = mimimi_allocate(allocator, total_animation_count * sizeof *animations);
	struct mimimi_image *images = mimimi_allocate(allocator, total_image_count * sizeof *images);
	
	set->standing = animations;
	set->standing_animation_count = animation_count;
	animations += animation_count;
	
	for (int i = 0 ; i < animation_count ; i++)
	{
		struct mimimi_animation *animation = set->standing + i;
		animation->count = image_count;
		animation->images = images;
		images += image_count;
		
		for (int i = 0 ; i < image_count ; i++)
		{
			struct mimimi_image *image = animation->images + i;
			image->width = width;
			image->height = height;
			image->colors = mimimi_allocate(allocator, width * height);
			for (int i = 0 ; i < width * height ; i++) image->colors[i] = 0;
		}
	}
	
	struct mimimi_animation *knocked = animations++;
	knocked->count = 1;
	knocked->images = images++;
	knocked->images[0].width = width;
	knocked->images[0].height = height;
	knocked->images[0].colors = mimimi_allocate(allocator, width * height);
	for (int i = 0 ; i < width * height ; i++) knocked->images[0].colors[i] = 0;
	
	struct mimimi_animation *falling = animations++;
	falling->count = 1;
	falling->images = images++;
	falling->images[0].width = width;
	falling->images[0].height = height;
	falling->images[0].colors = mimimi_allocate(allocator, width * height);
	for (int i = 0 ; i < width * height ; i++) falling->images[0].colors[i] = 0;
	
	set->knocked = knocked;
	set->falling = falling;
	
	for (int i = 0 ; i < width * height ; i++)
	{
		knocked->images[0].colors[i] = 0;
		falling->images[0].colors[i] = 0;
	}
	
	set->jumping = animations++;
	set->jumping->count = image_count;
	set->jumping->images = images;
	images += image_count;
	
	for (int i = 0 ; i < image_count ; i++)
	{
		struct mimimi_image *image = set->jumping->images + i;
		image->width = width;
		image->height = height;
		image->colors = mimimi_allocate(allocator, width * height);
		for (int i = 0 ; i < width * height ; i++) image->colors[i] = 0;
	}
}

static struct mimimi_appearance *mimimi_chapter_I_appearance(struct mimimi_allocator *allocator)
{
	struct mimimi_appearance *appearance = mimimi_allocate(allocator, sizeof *appearance);
	
	mimimi_chapter_I_animation_set(&appearance->left, allocator);
	mimimi_chapter_I_animation_set(&appearance->right, allocator);
	
	return appearance;
}

struct mimimi_chapter_I_sprite
{
	int x, y;
	struct mimimi_model *model;
};

static void mimimi_chapter_I_player(struct mimimi_behavior *compound, struct mimimi_chapter *chapter, struct mimimi_sprite *player, struct mimimi_history *history, struct mimimi_allocator *allocator)
{
	struct mimimi_allocator *compound_allocator = mimimi_compound_allocator(compound, allocator);
	
	struct mimimi_jump *jump = mimimi_allocate(compound_allocator, sizeof *jump);
	struct mimimi_behavior *jump_behavior = mimimi_jump(jump, player->physics, compound_allocator);
	
	struct mimimi_behavior *jump_choice = mimimi_choice(&player->physics->airborne, mimimi_noop, jump_behavior, compound_allocator);
	
	struct mimimi_controls *controls = mimimi_allocate(compound_allocator, sizeof *controls);
	mimimi_compound(compound, mimimi_controls(controls, player->walk, history, compound_allocator));;
	controls->jump = jump_choice;
}

struct mimimi_chapter_I_image
{
	int x, y, z;
	struct mimimi_image image;
};

struct mimimi_chapter_I_stage
{
	struct mimimi_ground *ground;
	int count;
	struct mimimi_chapter_I_image *images;
};

static void mimimi_chapter_I_add_image(struct mimimi_chapter_I_stage *stage, struct mimimi_image *image, int x, int y, int z, struct mimimi_allocator *allocator)
{
	stage->count++;
	stage->images = mimimi_reallocate(allocator, stage->images, stage->count * sizeof *stage->images);
	stage->images[stage->count - 1].image = *image;
	stage->images[stage->count - 1].x = x;
	stage->images[stage->count - 1].y = y;
	stage->images[stage->count - 1].z = z;
	for (int i = stage->count - 1 ; i > 0 ; i--)
	{
		int j = i - 1;
		if (stage->images[i].z < stage->images[j].z) break;
		struct mimimi_chapter_I_image image = stage->images[i];
		stage->images[i] = stage->images[j];
		stage->images[j] = image;
	}
}

#include "stage.c"

static void mimimi_chapter_I_box(struct mimimi_image *image, unsigned char stroke, unsigned char fill)
{
	int width = image->width;
	int height = image->height;
	
	unsigned char table[] = {0, stroke, fill};
	static unsigned char colors[10][10] =
	{
		{0,0,0,0,0,0,0,0,0,1},
		{0,0,0,0,0,0,0,1,1,1},
		{0,0,0,0,0,1,1,1,1,1},
		{0,0,0,0,1,1,1,1,2,2},
		{0,0,0,1,1,1,2,2,2,2},
		{0,0,1,1,1,2,2,2,2,2},
		{0,0,1,1,2,2,2,2,2,2},
		{0,1,1,1,2,2,2,2,2,2},
		{0,1,1,2,2,2,2,2,2,2},
		{1,1,1,2,2,2,2,2,2,2},
	};
	
	for (int y = 0 ; y < height ; y++)
	for (int x = 0 ; x < width ; x++)
		image->colors[x + y * width] = fill;
	
	for (int x = 0 ; x < width ; x++)
	{
		image->colors[x + 0 * width] = stroke;
		image->colors[x + 1 * width] = stroke;
		image->colors[x + (height - 1) * width] = stroke;
		image->colors[x + (height - 2) * width] = stroke;
	}
	
	for (int y = 0 ; y < height ; y++)
	{
		image->colors[0 + y * width] = stroke;
		image->colors[1 + y * width] = stroke;
		image->colors[(width - 1) + y * width] = stroke;
		image->colors[(width - 2) + y * width] = stroke;
	}
	
	for (int y0 = 0 ; y0 < 10 ; y0++)
	for (int x0 = 0 ; x0 < 10 ; x0++)
	{
		int x1 = width - 1 - x0;
		int y1 = height - 1 - y0;
		
		unsigned char color = table[colors[y0][x0]];
		
		image->colors[x0 + y0 * width] = color;
		image->colors[x1 + y0 * width] = color;
		image->colors[x0 + y1 * width] = color;
		image->colors[x1 + y1 * width] = color;
	}
}

static void mimimi_chapter_I_portrait(struct mimimi_image *image, struct mimimi_model *model, unsigned char background, struct mimimi_allocator *allocator)
{
	int width = image->width;
	int height = image->height;
	
	struct mimimi_pose_layer layer = {0, 42, 0};
	struct mimimi_pose pose = {width / 2, height / 2 + 28, 1, &layer};
	
	mimimi_pose(image, model, &pose, allocator);
	
	for (int i = 0 ; i < width * height ; i++)
	{
		if (image->colors[i] == 0)
			image->colors[i] = background;
	}
	
	static unsigned char mask[6][6] =
	{
		{0,0,0,0,0,0},
		{0,0,0,0,1,1},
		{0,0,0,1,1,1},
		{0,0,1,1,1,1},
		{0,1,1,1,1,1},
		{0,1,1,1,1,1},
	};
	
	for (int y0 = 0 ; y0 < 6 ; y0++)
	for (int x0 = 0 ; x0 < 6 ; x0++)
	{
		if (mask[y0][x0] != 0) continue;
		
		int x1 = width - 1 - x0;
		int y1 = height - 1 - y0;
		
		image->colors[x0 + y0 * width] = 0;
		image->colors[x1 + y0 * width] = 0;
		image->colors[x0 + y1 * width] = 0;
		image->colors[x1 + y1 * width] = 0;
	}
}

static void mimimi_chapter_I_stamp(struct mimimi_image *target, int x0, int y0, struct mimimi_image *source)
{
	for (int y1 = 0 ; y1 < source->height ; y1++)
	for (int x1 = 0 ; x1 < source->width ; x1++)
	{
		int x = x0 + x1;
		int y = y0 + y1;
		
		if (x < 0) continue;
		if (y < 0) continue;
		
		if (x >= target->width) break;
		if (y >= target->height) return;
		
		unsigned char color = source->colors[x1 + y1 * source->width];
		if (color == 0) continue;
		
		target->colors[x + y * target->width] = color;
	}
}

static int mimimi_chapter_I_word(char *text)
{
	int i = 0;
	while (text[i] != 0x20 && text[i] != 0) i++;
	return i;
}

static int mimimi_chapter_I_text(struct mimimi_image *image, struct mimimi_image *glyphs, int x, int y, int width, char *text, int measure, struct mimimi_allocator *allocator)
{
	int x1 = width + x;
	
	int sp = mimimi_measure_text(glyphs, " ");
	int current = x1 + 1;
	
	for (;;)
	{
		int i = mimimi_chapter_I_word(text);
		
		char *word = mimimi_allocate(allocator, i + 1);
		word[i] = 0;
		for (int j = 0 ; j < i ; j++) word[j] = text[j];
		
		int n = mimimi_measure_text(glyphs, word);
		
		if (current + n > x1)
		{
			y += 14;
			current = x;
			mimimi_deallocate(allocator, word);
			continue;
		}
		
		if (measure == 0)
			mimimi_text(image, glyphs, current, y, word);
		mimimi_deallocate(allocator, word);
		
		if (text[i] == 0) break;
		
		current += n + sp;
		
		text += i + 1;
	}
	
	return y;
}

static struct mimimi_image *mimimi_chapter_I_dialogue_box(struct mimimi_image *box, struct mimimi_image *glyphs, struct mimimi_model *model, char *name, int width, char *paragraph, struct mimimi_allocator *allocator)
{
	int height = 10 + mimimi_chapter_I_text(allocator->null, glyphs, 6, 20, width - 12, paragraph, 1, allocator);
	
	box->width = width;
	box->height = height;
	box->colors = mimimi_allocate(allocator, width * height);
	mimimi_chapter_I_box(box, 4, 13);
	
	struct mimimi_allocator *portrait_allocator = mimimi_temporary_allocator(allocator);
	
	struct mimimi_image *portrait = mimimi_allocate(portrait_allocator, sizeof *portrait);
	portrait->width = 18;
	portrait->height = 18;
	portrait->colors = mimimi_allocate(portrait_allocator, portrait->width * portrait->height);
	mimimi_chapter_I_portrait(portrait, model, 8, portrait_allocator);
	mimimi_chapter_I_stamp(box, 6, 6, portrait);
	
	mimimi_finish_temporary_allocator(portrait_allocator);
	
	int n = mimimi_text(box, glyphs, 28, 17, name);
	for (int i = 28 ; i < n - 1 ; i++)
		box->colors[i + 18 * box->width] = 4;
	
	mimimi_chapter_I_text(box, glyphs, 6, 20, width - 12, paragraph, 0, allocator);
	return box;
}

struct mimimi_chapter_I_paragraph
{
	struct mimimi_model *model;
	char *name;
	char *text;
};

struct mimimi_chapter_I_dialogue
{
	int count;
	struct mimimi_chapter_I_paragraph *paragraphs;
};

struct mimimi_chapter_I_strip_dialogue
{
	int count;
	void *texture;
	int *notches;
};

static struct mimimi_chapter_I_strip_dialogue *mimimi_chapter_I_strip_dialogue(struct mimimi_chapter_I_dialogue *dialogue, struct mimimi_image *glyphs, struct mimimi_engine *engine, struct mimimi_allocator *allocator)
{
	int *notches = mimimi_allocate(allocator, dialogue->count * sizeof *notches);
	struct mimimi_image *images = mimimi_allocate(allocator, dialogue->count * sizeof *images);
	int width = 128;
	int height = 0;
	
	for (int i = 0 ; i < dialogue->count ; i++)
	{
		struct mimimi_chapter_I_paragraph *paragraph = dialogue->paragraphs + i;
		mimimi_chapter_I_dialogue_box(images + i, glyphs, paragraph->model, paragraph->name, width, paragraph->text, allocator);
		height += images[i].height + 8;
		notches[i] = height;
	}
	
	struct mimimi_image *image = mimimi_allocate(allocator, sizeof *image);
	image->width = width;
	image->height = height;
	image->colors = mimimi_allocate(allocator, width * height);
	for (int i = 0 ; i < width * height ; i++)
		image->colors[i] = 0;
	
	mimimi_chapter_I_stamp(image, 0, 0, images);
	for (int i = 1 ; i < dialogue->count ; i++)
		mimimi_chapter_I_stamp(image, 0, notches[i - 1], images + i);
	
	mimimi_deallocate(allocator, images);
	
	void *texture = (*engine->texture)(engine->data, image);
	
	struct mimimi_chapter_I_strip_dialogue *strip_dialogue = mimimi_allocate(allocator, sizeof *strip_dialogue);
	strip_dialogue->count = dialogue->count;
	strip_dialogue->notches = notches;
	strip_dialogue->texture = texture;
	
	return strip_dialogue;
}

static void mimimi_chapter_I_show_strip_dialogue(struct mimimi_behavior *behavior, struct mimimi_engine *engine, struct mimimi_chapter *chapter, struct mimimi_chapter_I_strip_dialogue *strip_dialogue)
{
	int i = 0;
	int y = 0;
	int held = 1;
	
	while (i < strip_dialogue->count)
	{
		(*behavior->behave)(behavior->data);
		(*engine->stamp)(engine->data, 8, engine->size->height - y, strip_dialogue->texture);
		
		int distance = strip_dialogue->notches[i] - y;
		if (distance > 0) y += (distance - 1) / 8 + 1;
		
		if (chapter->left != 0 || chapter->right != 0)
		{
			if (held == 0) i++;
			held = 1;
		}
		else
		{
			held = 0;
		}
	}
	
	for (int i = 0 ; i < 32 ; i++)
	{
		(*behavior->behave)(behavior->data);
		(*engine->stamp)(engine->data, 8 - (128 + 8) * i / 32, engine->size->height - strip_dialogue->notches[strip_dialogue->count - 1], strip_dialogue->texture);
	}
}

struct mimimi_chapter_I
{
	struct mimimi_engine *engine;
	struct mimimi_chapter *chapter;
	struct mimimi_behavior *display;
};

static void mimimi_chapter_I_story(struct mimimi_behavior *behavior, struct mimimi_behavior *compound, void *data)
{
	struct mimimi_chapter_I *chapter_data = data;
	struct mimimi_engine *engine = chapter_data->engine;
	struct mimimi_chapter *chapter = chapter_data->chapter;
	struct mimimi_behavior *display = chapter_data->display;
	
	struct mimimi_allocator *compound_allocator = mimimi_compound_allocator(compound, engine->allocator);
	
	struct mimimi_ground *ground = mimimi_allocate(compound_allocator, sizeof *ground);
	ground->width = 64;
	ground->height = 64;
	ground->tiles = mimimi_allocate(compound_allocator, ground->width * ground->height);
	for (int i = 0 ; i < ground->width * ground->height ; i++)
		ground->tiles[i] = 1;
	
	struct mimimi_chapter_I_stage stage = {ground, 0, engine->allocator->null};
	mimimi_chapter_I_hideout(&stage, 1, 1, compound_allocator);
	
	struct mimimi_position *camera = mimimi_allocate(compound_allocator, sizeof *camera);
	
	struct mimimi_chapter_I_sprite infos[] =
	{
		{0, 0, mimimi_mirai_flash},
	};
	
	int count = sizeof infos / sizeof *infos;
	
	for (int y = 0 ; y < ground->height ; y++)
	for (int x = 0 ; x < ground->width ; x++)
	{
		if (ground->tiles[x + y * ground->width] == 2)
		{
			for (int i = 0 ; i < count ; i++)
			{
				infos[i].x += x;
				infos[i].y += y;
			}
			goto after;
		}
	}
	after:;
	
	struct mimimi_position **positions = mimimi_allocate(compound_allocator, count * sizeof *positions);
	mimimi_compound(compound, mimimi_collision(count, positions, 128, 256, 512, 64, engine->allocator));
	
	int *directions = mimimi_allocate(compound_allocator, count * sizeof *directions);
	
	struct mimimi_sprite **sprites = mimimi_allocate(engine->allocator, count * sizeof *sprites);
	struct mimimi_behavior *player_behavior = mimimi_compound_behavior(compound_allocator);
	
	struct mimimi_history *history = mimimi_allocate(compound_allocator, sizeof *history);
	mimimi_compound(compound, mimimi_history(history, chapter, compound_allocator));
	
	for (int i = 0 ; i < count ; i++)
	{
		struct mimimi_chapter_I_sprite *info = infos + i;
		
		struct mimimi_sprite *sprite = mimimi_sprite(ground, info->x * 128, info->y * 128, 80, 250, engine->allocator);
		mimimi_compound(compound, sprite->behavior);
		sprites[i] = sprite;
		
		if (i == 0)
		{
			mimimi_chapter_I_player(player_behavior, chapter, sprite, history, engine->allocator);
			mimimi_compound(player_behavior, mimimi_camera(camera, sprite->position, 0, -512, engine->allocator));
		}
		else
		{
			// todo
		}
		
		positions[i] = sprite->position;
	}
	
	struct mimimi_sprite *player = sprites[0];
	
	struct mimimi_image *overlay = mimimi_allocate(compound_allocator, sizeof *overlay);
	overlay->width = engine->size->width;
	overlay->height = engine->size->height;
	overlay->colors = mimimi_allocate(compound_allocator, overlay->width * overlay->height);
	
	for (int i = 0 ; i < overlay->width * overlay->height ; i++)
		overlay->colors[i] = 1;
	
	void *overlay_texture = (*engine->texture)(engine->data, overlay);
	mimimi_compound(display, mimimi_overlay(engine, overlay_texture, 0, 0, engine->allocator));
	
	int image_index = 0;
	
	while (image_index < stage.count)
	{
		struct mimimi_chapter_I_image *image = &stage.images[image_index];
		if (image->z < 0) break;
		void *texture = (*engine->texture)(engine->data, &image->image);
		mimimi_compound(display, mimimi_background(engine, camera, texture, image->x, image->y, image->z, engine->allocator));
		image_index++;
	}
	
	for (int i = 0 ; i < count ; i++)
	{
		struct mimimi_chapter_I_sprite *info = infos + i;
		struct mimimi_sprite *sprite = sprites[i];
		
		struct mimimi_appearance *appearance = mimimi_chapter_I_appearance(compound_allocator);
		mimimi_appearance(appearance, info->model, 24, 48, engine->allocator);
		
		struct mimimi_video_appearance *video_appearance = mimimi_allocate(compound_allocator, sizeof *video_appearance);
		mimimi_video_appearance(video_appearance, appearance, engine, compound_allocator);
		
		mimimi_compound(display, mimimi_display(engine, camera, sprite, video_appearance, directions + i, 24, 48, engine->allocator));
	}
	
	mimimi_deallocate(engine->allocator, sprites);
	
	while (image_index < stage.count)
	{
		struct mimimi_chapter_I_image *image = &stage.images[image_index];
		void *texture = (*engine->texture)(engine->data, &image->image);
		mimimi_compound(display, mimimi_background(engine, camera, texture, image->x, image->y, image->z + 1, engine->allocator));
		image_index++;
	}
	
	struct mimimi_image *glyphs = mimimi_allocate(compound_allocator, mimimi_font_image_count * sizeof *glyphs);
	unsigned char *colors = mimimi_allocate(compound_allocator, mimimi_font_color_count);
	
	for (int i = 0 ; i < mimimi_font_image_count ; i++)
		glyphs[i].colors = colors + i * mimimi_font_image_color_count;
	
	mimimi_glyphs(glyphs, mimimi_font, 4);
	
	struct mimimi_chapter_I_paragraph paragraphs[] =
	{
		{mimimi_mirai_flash, "Mirai", "There seems to be nothing for me to do. Hmmm... Maybe I can find something to do later."},
		{mimimi_ryouna_flash, "Ryouna", "Hear that? If you come back later, then maybe you can find something to do!"},
		{mimimi_ryoubi_flash, "Ryoubi", "Who are you talking with? Are you stupid or something?"},
		{mimimi_ryouna_flash, "Ryouna", "Nyaa~ Ryoubi, I really like when you yell insults at me!"},
		{mimimi_mirai_flash, "Mirai", "Wha-? What are you two doing here?"},
		{mimimi_haruka_flash, "Haruka", "Hehe, don't worry about it!"},
	};
	
	struct mimimi_chapter_I_dialogue dialogue = {sizeof paragraphs / sizeof *paragraphs, paragraphs};
	
	struct mimimi_chapter_I_strip_dialogue *strip_dialogue = mimimi_chapter_I_strip_dialogue(&dialogue, glyphs, engine, compound_allocator);
	
	while (positions[0]->x > 512)
	{
		(*player_behavior->behave)(player_behavior->data);
		(*behavior->behave)(behavior->data);
	}
	
	player->walk->direction = 0;
	mimimi_chapter_I_show_strip_dialogue(behavior, engine, chapter, strip_dialogue);
	
	for (;;)
	{
		(*behavior->behave)(behavior->data);
		(*player_behavior->behave)(player_behavior->data);
	}
}

struct mimimi_chapter *mimimi_chapter_I(struct mimimi_engine *engine)
{
	struct mimimi_behavior *compound = mimimi_compound_behavior(engine->allocator);
	struct mimimi_allocator *compound_allocator = mimimi_compound_allocator(compound, engine->allocator);
	
	struct mimimi_chapter *chapter = mimimi_allocate(compound_allocator, sizeof *chapter);
	chapter->behavior = compound;
	
	struct mimimi_behavior *display = mimimi_compound_behavior(engine->allocator);
	mimimi_compound(compound, display);
	
	struct mimimi_chapter_I *data = mimimi_allocate(compound_allocator, sizeof *data);
	data->engine = engine;
	data->chapter = chapter;
	data->display = display;
	
	mimimi_coroutine(engine->coroutines, compound, &mimimi_chapter_I_story, data, compound_allocator);
	return chapter;
}
